type 'a types = Base of 'a | Curry of 'a types * 'a types

(* Var are bounded variables, Const are context constants or constructors 
   (Zero, Suc for instance), Unk are free variables with type in form 
   (args types, base type), Rec_const are recursive constant (nat_rec for
   instance)
 *)
type 'a term = 
    Var of int
  | Const of string
  | Abs of 'a term
  | App of 'a term * 'a term
  | Unk of string * ('a types list * 'a types)
  | Rec_const of string

(* more accurate form for terms: recursive head normal form *)
type 'a head_normal_form =
    { n : int;
      head : 'a term;
      args : 'a term list;
      p : int;
      alphas : 'a types list;
      beta : 'a types;
      t : 'a types }


(* table containing types of the constants *)
let type_table : (string, string types) Hashtbl.t = Hashtbl.create 10

let add_type = Hashtbl.add type_table

let find_type t = 
  try Hashtbl.find type_table t
  with Not_found -> failwith ("untyped constant: "^t)


(* table containing constructors associated with recursive constants *)
let constructors_table : (string, string head_normal_form list) Hashtbl.t 
    = Hashtbl.create 10

let add_constructors = Hashtbl.add constructors_table

let find_constructors t = 
  try Hashtbl.find constructors_table t
  with Not_found -> failwith ("unknown constructor: "^t)


(* table containing base type associated with recursive constants *)
let base_types_table : (string, string types) Hashtbl.t = Hashtbl.create 10
    
let add_base_types = Hashtbl.add base_types_table
    
let find_base_types t = 
  try Hashtbl.find base_types_table t
  with Not_found -> failwith ("unknown constructor: "^t)
                

(* this execption is raised when an expression of type 'a head_normal_form 
   does not represent a term in recursive head normal form (typically when
   the head is actually an abstraction)
*) 
exception Not_normal_form of string term * string types




(* Representation of types with lists: *)
(* uncurryfy ('b1 -> 'b2 -> ... -> 'bn -> 'a) = ['b1, ..., 'bn] 
   if 'a is a basis type *)
let rec uncurryfy = function
  | Curry(a, b) -> a::(uncurryfy b)
  | b -> []

(* rev_uncurryfy ('b1 -> 'b2 -> ... -> 'bn -> 'a) = ['bn, ..., 'b1], 'a
   if 'a is a basis type *)
let rev_uncurryfy t =
  let rec aux accu = function
    | Curry(a, b) -> aux (a::accu) b
    | b -> accu, b
  in aux [] t

(* rev_curryfy 'a ['bn, ..., 'b1] = ('b1 -> 'b2 -> ... -> 'bn -> 'a) *)
let rec rev_curryfy b = function
  | x::q -> rev_curryfy (Curry (x, b)) q
  | [] -> b


(* creation of new variables *)
let unk_counter = ref (-1)

let new_unk (l, a) = 
  let rec aux l = function
    | Curry (a, b) -> aux (a::l) b
    | b -> l, b
  in 
  unk_counter := !unk_counter + 1; 
  if (!unk_counter = max_int) then 
    failwith "Too many variable created!"
  else
    Unk (string_of_int !unk_counter, aux l a)


(* term to head_normal_form, s must be in long recursive head normal form *)
let term_to_hnf (s, t) = 
  let rec apps l i = function
    | App (a, b) -> apps (b::l) (i+1) a
    | Abs _ -> 	
	raise (Not_normal_form (s, t))
    | c -> c, l, i
  in
  let rec abs alphas n = function
    | Abs s, Curry (a, b) -> abs (a::alphas) (n+1) (s, b)
    | Abs _, b -> 
	raise (Not_normal_form (s, t))
    | c, b -> 
	let h, args, p = apps [] 0 c in
	{ n = n; head = h; args = args; p = p; alphas = alphas; 
	  beta = b; t = t}
  in
  let _::q = [t] in  (* without this, polymorph type can't be generalized *)
  abs q 0 (s, t)


(* abs r n abstract the term r n times *)
let rec abs r = function
  | 0 -> r
  | n -> abs (Abs r) (n-1)

(* inverse function of term_to_hnf *)
let hnf_to_term e = 
  let rec app c = function
    | [] -> c
    | ei::es -> app (App (c, ei)) es
  in abs (app e.head e.args) e.n, e.t

(* type of the head of an term in rhnf representation *)
let hnf_head_types h = 
  match h.head with
  | Var j -> List.nth h.alphas j
  | Const s -> find_type s
  | Rec_const s -> find_type s
  | Unk (_, (a, b)) -> rev_curryfy b a
  | _ -> 
      raise (let r = hnf_to_term h in Not_normal_form (fst r, snd r))
