open Defs

(* implementation of a rewriting system *)
exception Not_match

(* look if s can match t, return the substitution in that case, raise 
   Not_match otherwise *)
let match_on_top s t = 
  let rec ind = function
    | Abs s, Abs t -> ind (s, t)
    | App (s, t), App (u, v) -> List.rev_append (ind (s, u)) (ind (t, v))
    | Unk (i, _), c -> [i, c]
    | s, t -> if s = t then [] else raise Not_match
  in
  let rec check accu = function
    | [] -> accu
    | (i, c)::q -> 
	try
	  if List.assoc i q = c then check accu q
	  else raise Not_match
	with Not_found -> check ((i, c)::accu) q
  in
  check [] (ind (s, t))

(* apply the substitution \mymath{\theta} *)
let rec apply_subst theta = function
  | Abs s -> Abs (apply_subst theta s)
  | App (s, t) -> App (apply_subst theta s, apply_subst theta t)
  | Unk (f, _) as s -> (
      try
	List.assoc f theta 
      with Not_found -> s )
  | s -> s

(* rewrite the term t on top using rule s \mymath{\rewrite} t *)
let rec rewrite_on_top (s, s') t =
  apply_subst (match_on_top s t) s'

(* the same, with a system of rewriting rules *)
let rewrite_system_on_top sss t =
  let rec aux = function
    | [] -> raise Not_match
    | x::s -> 
	try 
	  rewrite_on_top x t
	with 
	  Not_match -> aux s 
  in
  aux sss

(* rewriting in only to the left part of the terms, using an external
   function in addition (typically the \mymath\beta-reduction) 
*)
let rewrite_system_on_left_func f sss t =
  let rec ind args t = 
    try rewrite_system_on_top sss t
    with Not_match ->
      match t with
    | Abs u -> Abs (ind [] u)
    | App (u, v) -> App (ind (v::args) u, v)
    | Rec_const s as u -> 
	let n = List.length (find_constructors s) in
	let rec app_args c = function 
	  | 0, x::_ -> App(c, ind [] x)
	  | n, x::q -> app_args (App(c, x)) (n-1, q)
	  | _ -> failwith "function asn't enough arguments"
	in
	abs (app_args u (n, args)) (n+1)
    | c -> raise Not_match
  in
  let rec rew_loop m =
    try
      rew_loop (f (ind [] m))
    with Not_match -> m 
  in
  rew_loop (f t)


(* \mymath{\beta}-reductions: *)
let rec shift i d = function
  | Var j when j >= d -> Var (j+i)
  | Abs s -> Abs (shift i (d+1) s)
  | App (s, t) -> App (shift i d s, shift i d t)
  | s -> s

let replace t =
  let rec aux i = function
    | Var j -> 
	if j = i then shift i 0 t 
	else (if j < i then Var j else Var (j-1))
    | App (u, v) -> App ((aux i u), (aux i v))
    | Abs u -> Abs (aux (i + 1) u)
    | t -> t
  in aux 0

(* \mymath\beta-head normal form *)
let rec beta_hnf = function
  | Abs s -> Abs (beta_hnf s)
  | App (s, t) -> (
      match beta_hnf s with
      | Abs u -> beta_hnf (replace t u)
      | u -> App (u, t) )
  | t -> t

(* \mymath\beta-normal form *)
let beta_nf t = 
  let rec aux = function
    | Abs s -> Abs (aux s)
    | App (s, t) -> App (aux s, aux (beta_hnf t))
    | t -> t
  in aux (beta_hnf t)
      
