open Defs
open Printer

exception F  (* failure of the unification *)

type 'a subst = string * 'a term 

type 'a request = 'a head_normal_form * 'a head_normal_form  

type 'a result_tree = 
    'a subst list list


(* printing primitives *)
let print_constraints l =  
   List.iter
    (fun a -> Queue.iter (fun (e1, e2) ->
      parse_print (fst(hnf_to_term e1));
      print_string ", ";
      pretty_print (fst(hnf_to_term e2));
      print_string "; ";) a;
      print_string " | "
    )
    l

let print_subst theta =
  List.iter
    (fun (s, t) ->
      print_Ind s;
      print_string " -> ";
      pretty_print t;
      print_string "; ";
    )
    theta

let print_terms print_func theta =
  List.iter
    (fun t ->
      print_func t;
      print_string "; ";
    )
    theta

let rec print_dec = function
  | 0 -> ()
  | n -> print_char '>'; print_dec (n-1)



(* return the rec-flexible argument of the term e (in rhnf) *)
let rec rec_flexible_arg e rules = 
   match e.head with
  | Rec_const s -> 
      let n = List.length (find_constructors s) in
      let r = List.nth e.args n in
      let h = term_to_hnf (r, find_base_types s) in
      rec_flexible_arg h rules
  | s -> (fst (hnf_to_term e))
	

(* apply the variable represented by the indixes 0 to k - 1 to c *)
let rec apply_var c = function
  | 0 -> c
  | k -> apply_var (App(c, Var (k - 1))) (k-1)


(* the SIMPL procedure *)
let simpl n rules =
  let rec simpl n accu =
    let rec dec q n ns = function
    | [], [], [] -> q
    | e1 :: l1, e2 :: l2, t::ts -> 
	dec (Queue.push q (term_to_hnf
   (Rewriting.rewrite_system_on_left_func Rewriting.beta_nf rules (abs e1 n),
    rev_curryfy t ns), 
			   term_to_hnf 
   (Rewriting.rewrite_system_on_left_func Rewriting.beta_nf rules (abs e2 n), 
    rev_curryfy t ns))) 
	  n ns (l1, l2, ts)
    | _ -> raise F
    in
    if Queue.is_empty n then accu 
    else
      let (e1, e2), n' = Queue.pop n in 
      match e1.head, e2.head with
      | Var i, Var j -> 
	  if i = j 
	  then 
	    simpl (dec n' e1.n e1.alphas 
		     (e1.args, e2.args, uncurryfy (hnf_head_types e1)))
	      accu
	  else raise F
      | Const s, Const t -> 
	  if s = t 
	  then 
	    simpl (dec n' e1.n e1.alphas 
		     (e1.args, e2.args, uncurryfy (hnf_head_types e1)))
	      accu
	  else raise F
      | Unk (_, _), Unk (_, _) -> simpl n' accu (* not necessary actually *)
      | Unk (_, _), _ -> simpl n' (Queue.push accu (e1, e2))
      | _, Unk (_, _) -> simpl (Queue.add_as_first n' (e2, e1)) accu (* idem *)
      | Rec_const _, _ -> simpl n' (Queue.push accu (e1, e2))
      | _, Rec_const _ -> simpl (Queue.add_as_first n' (e2, e1)) accu (* idem *)
      | _ -> raise F
  in Queue.pop (simpl n (Queue.create ()))


(* the MATCH procedure *)
(* imitation rule *)
let rec imitation cl (e1, e2) theta =
  match e1.head with
  | Unk (f, (l, t)) ->   
      let rec apply_new_unk c = function
	| Curry (a, b) -> 
	    apply_new_unk (App(c, apply_var (new_unk (l, a)) e1.p)) 
	      b
	| _ -> c
      in
      ((f, abs (apply_new_unk e2.head (hnf_head_types e2)) e1.p), cl)::theta
(* imitation in case of a rec_flexible term: none (see example \ref{no_ext_imit}) *)
  | Rec_const s -> 
      theta
  | _ -> pretty_print (fst (hnf_to_term e1));
      print_newline ();
      pretty_print (fst (hnf_to_term e2));
      print_newline ();
      failwith "Not in simplified form (flex-rigid or rec_flex-rigid)"

(* projection rule *)
let proj rules cl (e1i, e2i) theta =
let rec proj cl (e1, e2) theta =
  match e1.head with
  | Unk (f, (l, beta)) ->   
      let rec aux accu i = function
	| [] -> accu
	| ai::q -> let rec apply_new_unk c = function
	    | Curry (a, b) -> apply_new_unk (App (c, apply_var (new_unk (l, a)) e1.p)) b
	    | b -> 
		if b = beta 
		then ((f, abs c e1.p), cl) :: accu
		else accu
	in aux (apply_new_unk (Var i) ai) (i+1) q
      in aux theta 0 l
(* projection in case of a rec_flexible term *)
  | Rec_const s -> 
      proj cl (term_to_hnf (rec_flexible_arg e1 rules, find_base_types s), e2) theta 
  | _ -> 
      failwith "Not in simplified form (flex-rigid or rec_flex-rigid)"
in
proj cl (e1i, e2i) theta


(* recursion rule *)
(* Constants are given in form 2-uples:
    - const: name of the const
    - typ: types of the other arguments of the constant, possibly polymorph 
   The set of the recursiv contant we can use is given as a queue
*)
let rec_const cl (e1, e2) theta ((const, typ), nb) =
  if nb <= 0 then theta else (
  let cl = Queue.push cl ((const, typ), nb-1) in
  match e1.head with
  | Unk (f, (l, beta)) ->   
      let rec apply_new_unk c k = function
	| a::b -> apply_new_unk (App (c, apply_var (new_unk (l, a)) k)) k b
	| [] -> c
      in
      ((f, abs (apply_var (apply_new_unk (Rec_const const)
		  e1.p (typ (rev_curryfy beta l))) e1.p) e1.p), cl)
      :: theta
(* we don't use a rec_const iterativly, but to do so will be equivalent 
   (see example \ref{no_ext_rec}) *)
  | Rec_const s -> theta
  | _ -> failwith "Not in simplified form (flex-rigid or rec_flex-rigid)"
 ) 


(* apply the substitutions \mymath{?F \mapsto t} to the constraints n, 
   and perform rewriting rules *)
let apply_subst (f, t) n system =
  let rec aux = function
    | Abs s -> Abs (aux s)
    | App (s, t) -> App (aux s, aux t) 
    | Unk (g, _) when f = g -> t
    | c -> c
  in
  let rec aux2 h = 
    term_to_hnf 
      (
       let e, ty = (hnf_to_term h) in 
       Rewriting.rewrite_system_on_left_func Rewriting.beta_nf system (aux e),
       ty )
  in
  Queue.map
    (fun (e1, e2) -> aux2 e1, aux2 e2)
    n
  
let rec substs_in_term substs = function
  | Abs s -> Abs (substs_in_term substs s)
  | App (u, v) -> App (substs_in_term substs u, substs_in_term substs v)
  | Unk (j, _) as s -> (
      try 
	List.assoc j substs 
      with Not_found -> s
     )
  | s -> s


(* unification procedure *)
let unif_consts consts rules n prof punc = 
  let rec unif_consts cl (es, n) prof =
    if prof <= 0 then []
    else 
      let subst =
	let rec fold cl theta = function
	  | [] -> theta
	  | c :: q -> let _, cl = Queue.pop cl in
	    fold (Queue.push cl c) 
	      (rec_const cl es theta c)
	      q
	in
	let qcl = Queue.from_list cl in
	fold qcl
	  (proj rules qcl es (imitation qcl es []))
	  cl
      in
      (match subst with [] -> [] (* no subsitution possible, we have failed *)
      | _ -> 

	    (List.fold_left
	       (fun l ((s, t), cl) -> 
		 List.rev_append 
		   (List.rev_map 
		      (fun substs -> (s, substs_in_term substs t)::substs)
		      (try
			unif_consts (Queue.to_list cl)
			  (simpl (apply_subst (s, t) (Queue.push n es) rules) rules) 
			  (prof-1)
		      with 
		      | F -> []
		      | Queue.Empty -> [[]] 
                          (* Success, we return the empty substitution *)
		      )
		   )
		   l
	       )
	       []
	       subst
	    ) ) 
  in 
  try
    unif_consts consts (simpl n rules) prof
  with
  | F -> []
  | Queue.Empty -> [[]]


(* get the final substitution from the search tree *)
let final_subst s t =
  List.map (fun l -> Rewriting.beta_nf (List.assoc s l) ) t
